/**
 * Copyright (c) 2008-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.jaas;

import java.io.File;
import java.io.IOException;
import java.security.Principal;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Logger;

import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.NameCallback;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.callback.UnsupportedCallbackException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

/**
 * A basic login module implementation that use a properties file for users and
 * groups definition.
 * 
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since 1.0
 * 
 */
public class PropertiesLoginModule implements LoginModule {

    /**
     * The property used to define the user property file
     */
    private static final String USER_FILE = "org.ow2.petals.security.properties.user";

    /**
     * The property used to define the group property file
     */
    private static final String GROUP_FILE = "org.ow2.petals.security.properties.group";

    private static final String DEBUG = "debug";

    /**
     * Subject cache from the
     * {@link #initialize(Subject, CallbackHandler, Map, Map)}
     */
    private Subject subject;

    /**
     * Handler cache from
     * {@link #initialize(Subject, CallbackHandler, Map, Map)}
     */
    private CallbackHandler callbackHandler;

    /**
     * The users properties file name to be used to load users
     */
    private String usersFile;

    /**
     * The groups properties file name to be used to load groups
     */
    private String groupsFile;

    /**
     * Properties loaded from the users file
     */
    private final Properties users;

    /**
     * Properties loaded from the groups file
     */
    private final Properties groups;

    /**
     * USer to authenticate
     */
    private String user;

    /**
     * Set of principals
     */
    private final Set<Principal> principals;

    /**
     * The base dir where configuration files are available
     */
    private File baseDir;

    private boolean debug;

    /**
     * 
     */
    private static final Logger logger = Logger.getLogger(PropertiesLoginModule.class.getName());

    /**
     * Creates a new instance of PropertiesLoginModule
     */
    public PropertiesLoginModule() {
        this.users = new Properties();
        this.groups = new Properties();
        this.principals = new HashSet<Principal>();
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.security.auth.spi.LoginModule#initialize(javax.security.auth.Subject,
     *      javax.security.auth.callback.CallbackHandler, java.util.Map,
     *      java.util.Map)
     */
    public void initialize(Subject subject, CallbackHandler callbackHandler,
            Map<String, ?> sharedState, Map<String, ?> options) {
        logger.info("Initializing login module");
        this.subject = subject;
        this.callbackHandler = callbackHandler;

        if (System.getProperty("java.security.auth.login.config") != null) {
            // get the base dir from the system property. This property define
            // the login module property file to be used.
            this.baseDir = new File(System.getProperty("java.security.auth.login.config"))
                    .getParentFile();
        } else {
            // FIXME : PETALS ROOT PATH
            // Should not happen...
            this.baseDir = new File(".");
        }

        // get properties which have been defined in the login module file
        this.usersFile = options.get(PropertiesLoginModule.USER_FILE).toString();
        this.groupsFile = options.get(PropertiesLoginModule.GROUP_FILE).toString();
        Object tmp = options.get(PropertiesLoginModule.DEBUG);
        if (tmp == null) {
            tmp = Boolean.FALSE.toString();
        }
        this.debug = Boolean.valueOf(tmp.toString());

        if (debug) {
            logger.info("Base dir : " + this.baseDir.getAbsolutePath());
            logger.info("Users file : " + this.usersFile + ", Groups file : " + this.groupsFile);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.security.auth.spi.LoginModule#login()
     */
    public boolean login() throws LoginException {
        if (debug) {
            logger.info("Login");
        }

        // load users
        File file = new File(this.baseDir, this.usersFile);
        try {
            this.users.load(new java.io.FileInputStream(file));
        } catch (final IOException ioe) {
            throw new LoginException("Unable to load user properties file "
                    + file.getAbsolutePath());
        }

        // load groups
        file = new File(this.baseDir, this.groupsFile);
        try {
            this.groups.load(new java.io.FileInputStream(file));
        } catch (final IOException ioe) {
            throw new LoginException("Unable to load group properties file "
                    + file.getAbsolutePath());
        }

        // create and handle callbacks (fill them)
        final Callback[] callbacks = new Callback[2];
        callbacks[0] = new NameCallback("Username: ");
        callbacks[1] = new PasswordCallback("Password: ", false);
        try {
            this.callbackHandler.handle(callbacks);
        } catch (final IOException ioe) {
            throw new LoginException(ioe.getMessage());
        } catch (final UnsupportedCallbackException uce) {
            throw new LoginException(uce.getMessage()
                    + " not available to obtain information from user");
        }

        // Get and check password. The user has been filled by the callback, its
        // password too...
        this.user = ((NameCallback) callbacks[0]).getName();
        char[] tmpPassword = ((PasswordCallback) callbacks[1]).getPassword();
        if (tmpPassword == null) {
            // no password is empty one
            tmpPassword = new char[0];
        }

        // get the password defined in the users properties file
        final String password = this.users.getProperty(this.user);
        if (password == null) {
            throw new FailedLoginException("User " + this.user + " does not exist");
        }
        String tmpPasswordString = new String(tmpPassword);
        if (!password.equals(tmpPasswordString)) {
            throw new FailedLoginException("Password does not match, file one is '" + password
                    + "', retrieved one is '" + tmpPasswordString + "'");
        }

        // clean the users
        this.users.clear();

        if (debug) {
            logger.info("User '" + this.user + "' is logged in");
        }
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.security.auth.spi.LoginModule#commit()
     */
    public boolean commit() throws LoginException {
        if (debug) {
            logger.fine("Commit");
        }

        // user is logged in, create its Principal
        this.principals.add(new UserPrincipal(this.user));

        Enumeration<?> groupsKeys = groups.keys();
        while (groupsKeys.hasMoreElements()) {
            final String groupName = (String) groupsKeys.nextElement();
            final String[] groupUsers = ((String) this.groups.getProperty(groupName)).split(",");
            for (final String userInGroup : groupUsers) {
                if (this.user.equals(userInGroup)) {
                    // user found in group, create a Principal
                    this.principals.add(new GroupPrincipal(groupName));
                    break;
                }
            }
        }

        // add all the principals to the subject
        this.subject.getPrincipals().addAll(this.principals);
        this.clean();

        if (debug) {
            logger.fine("Commit done");
        }
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.security.auth.spi.LoginModule#abort()
     */
    public boolean abort() throws LoginException {
        if (debug) {

            logger.fine("Abort");
        }
        this.clean();
        logger.fine("Abort done");
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.security.auth.spi.LoginModule#logout()
     */
    public boolean logout() throws LoginException {
        logger.fine("Logout");
        this.subject.getPrincipals().removeAll(this.principals);
        this.principals.clear();
        logger.fine("Logout done");
        return true;
    }

    private void clean() {
        logger.fine("Clean");
        this.groups.clear();
        this.user = null;
        logger.fine("Clean done");
    }
}
